@@ -1,6 +1,6 @@
 from pyrpm.spec import Spec, replace_macros
 
-spec = Spec.from_file('llvm.spec')
+spec = Spec.from_file("llvm.spec")
 print(spec.version)  # 3.8.0
 print(spec.sources[0])  # http://llvm.org/releases/%{version}/%{name}-%{version}.src.tar.xz
 print(replace_macros(spec.sources[0], spec))  # http://llvm.org/releases/3.8.0/llvm-3.8.0.src.tar.xz
