@@ -10,22 +10,34 @@ add support for the missing pieces.
 
 """
 
+import os
 import re
+import sys
+from warnings import warn
 from abc import ABCMeta, abstractmethod
+from typing import Any, Dict, List, Optional, Union, Tuple, Type, cast
 
-__all__ = ["Spec", "replace_macros", "Package"]
+if sys.version_info < (3, 7):
+    re.Pattern = Any
+    re.Match = Any
+
+__all__ = ["Spec", "replace_macros", "Package", "warnings_enabled"]
+
+
+# Set this to True if you want the library to issue warnings during parsing.
+warnings_enabled: bool = False
 
 
 class _Tag(metaclass=ABCMeta):
-    def __init__(self, name, pattern_obj, attr_type):
+    def __init__(self, name: str, pattern_obj: re.Pattern, attr_type: Type[Any]) -> None:
         self.name = name
         self.pattern_obj = pattern_obj
         self.attr_type = attr_type
 
-    def test(self, line):
+    def test(self, line: str) -> Optional[re.Match]:
         return re.search(self.pattern_obj, line)
 
-    def update(self, spec_obj, context, match_obj, line):
+    def update(self, spec_obj: "Spec", context: Dict[str, Any], match_obj: re.Match, line: str) -> Any:
         """Update given spec object and parse context and return them again.
 
         :param spec_obj: An instance of Spec class
@@ -47,7 +59,7 @@ class _Tag(metaclass=ABCMeta):
         pass
 
     @staticmethod
-    def current_target(spec_obj, context):
+    def current_target(spec_obj: "Spec", context: Dict[str, Any]) -> Union["Spec", "Package"]:
         target_obj = spec_obj
         if context["current_subpackage"] is not None:
             target_obj = context["current_subpackage"]
@@ -57,10 +69,13 @@ class _Tag(metaclass=ABCMeta):
 class _NameValue(_Tag):
     """Parse a simple name → value tag."""
 
-    def __init__(self, name, pattern_obj, attr_type=None):
-        super().__init__(name, pattern_obj, attr_type if attr_type else str)
+    def __init__(self, name: str, pattern_obj: re.Pattern, attr_type: Optional[Type[Any]] = None) -> None:
+        super().__init__(name, pattern_obj, cast(Type[Any], attr_type if attr_type else str))
+
+    def update_impl(self, spec_obj: "Spec", context: Dict[str, Any], match_obj: re.Match, line: str) -> Tuple["Spec", dict]:
+        if self.name == "changelog":
+            context["current_subpackage"] = None
 
-    def update_impl(self, spec_obj, context, match_obj, line):
         target_obj = _Tag.current_target(spec_obj, context)
         value = match_obj.group(1)
 
@@ -69,10 +84,43 @@ class _NameValue(_Tag):
             spec_obj.packages = []
             spec_obj.packages.append(Package(value))
 
-        setattr(target_obj, self.name, self.attr_type(value))
+        if self.name in ["description", "changelog"]:
+            context["multiline"] = self.name
+        else:
+            setattr(target_obj, self.name, self.attr_type(value))
+
         return spec_obj, context
 
 
+class _SetterMacroDef(_Tag):
+    """Parse global macro definitions."""
+
+    def __init__(self, name: str, pattern_obj: re.Pattern) -> None:
+        super().__init__(name, pattern_obj, str)
+
+    def get_namespace(self, spec_obj, context):
+        raise NotImplementedError()
+
+    def update_impl(self, spec_obj: "Spec", context: Dict[str, Any], match_obj: re.Match, line: str) -> Tuple["Spec", dict]:
+        name, value = match_obj.groups()
+        setattr(self.get_namespace(spec_obj, context), name, str(value))
+        return spec_obj, context
+
+
+class _GlobalMacroDef(_SetterMacroDef):
+    """Parse global macro definitions."""
+
+    def get_namespace(self, spec_obj: "Spec", context: Dict[str, Any]) -> "Spec":
+        return spec_obj
+
+
+class _LocalMacroDef(_SetterMacroDef):
+    """Parse define macro definitions."""
+
+    def get_namespace(self, spec_obj: "Spec", context: Dict[str, Any]) -> "Spec":
+        return context["current_subpackage"]
+
+
 class _MacroDef(_Tag):
     """Parse global macro definitions."""
 
@@ -91,21 +139,21 @@ class _MacroDef(_Tag):
 class _List(_Tag):
     """Parse a tag that expands to a list."""
 
-    def __init__(self, name, pattern_obj):
+    def __init__(self, name: str, pattern_obj: re.Pattern) -> None:
         super().__init__(name, pattern_obj, list)
 
-    def update_impl(self, spec_obj, context, match_obj, line):
+    def update_impl(self, spec_obj: "Spec", context: Dict[str, Any], match_obj: re.Match, line: str) -> Tuple["Spec", dict]:
         target_obj = _Tag.current_target(spec_obj, context)
 
         if not hasattr(target_obj, self.name):
-            setattr(target_obj, self.name, list())
+            setattr(target_obj, self.name, [])
 
         value = match_obj.group(1)
         if self.name == "packages":
             if value == "-n":
                 subpackage_name = line.rsplit(" ", 1)[-1].rstrip()
             else:
-                subpackage_name = "{}-{}".format(spec_obj.name, value)
+                subpackage_name = f"{spec_obj.name}-{value}"
             package = Package(subpackage_name)
             context["current_subpackage"] = package
             package.is_subpackage = True
@@ -117,6 +165,8 @@ class _List(_Tag):
             "obsoletes",
             "provides",
         ]:
+            # Remove comments on same line
+            value = value.split("#", 2)[0].rstrip()
             # Macros are valid in requirements
             value = replace_macros(value, spec=spec_obj)
 
@@ -127,7 +177,7 @@ class _List(_Tag):
 
             # 1. Tokenize
             tokens = [val for val in re.split("[\t\n, ]", value) if val != ""]
-            values = []
+            values: List[str] = []
 
             # 2. Join
             add = False
@@ -152,33 +202,69 @@ class _List(_Tag):
 class _ListAndDict(_Tag):
     """Parse a tag that expands to a list and to a dict."""
 
-    def __init__(self, name, pattern_obj):
+    def __init__(self, name: str, pattern_obj: re.Pattern) -> None:
         super().__init__(name, pattern_obj, list)
 
-    def update_impl(self, spec_obj, context, match_obj, line):
+    def update_impl(self, spec_obj: "Spec", context: Dict[str, Any], match_obj: re.Match, line: str) -> Tuple["Spec", dict]:
         source_name, value = match_obj.groups()
-        dictionary = getattr(spec_obj, "{}_dict".format(self.name))
+        dictionary = getattr(spec_obj, f"{self.name}_dict")
         dictionary[source_name] = value
         target_obj = _Tag.current_target(spec_obj, context)
         getattr(target_obj, self.name).append(value)
         return spec_obj, context
 
 
+class _SplitValue(_NameValue):
+    """Parse a (name->value) tag, and at the same time split the tag to a list."""
+
+    def __init__(self, name: str, pattern_obj: re.Pattern, sep: str = None) -> None:
+        super().__init__(name, pattern_obj)
+        self.name_list = f"{name}_list"
+        self.sep = sep
+
+    def update_impl(self, spec_obj: "Spec", context: Dict[str, Any], match_obj: re.Match, line: str) -> Tuple["Spec", dict]:
+        super().update_impl(spec_obj, context, match_obj, line)
+
+        target_obj = _Tag.current_target(spec_obj, context)
+        value = getattr(target_obj, self.name)
+        value = value.split(self.sep)
+        setattr(target_obj, self.name_list, value)
+
+        return spec_obj, context
+
+
 def re_tag_compile(tag):
     return re.compile(tag, re.IGNORECASE)
 
 
+class _DummyMacroDef(_Tag):
+    """Parse global macro definitions."""
+
+    def __init__(self, name, pattern_obj):
+        super().__init__(name, pattern_obj, str)
+
+    def update_impl(self, spec_obj, context, _, line):
+        context["line_processor"] = None
+        if warnings_enabled:
+            warn("Unknown macro: " + line)
+        return spec_obj, context
+
+
 _tags = [
     _NameValue("name", re_tag_compile(r"^Name\s*:\s*(\S+)")),
     _NameValue("version", re_tag_compile(r"^Version\s*:\s*(\S+)")),
     _NameValue("epoch", re_tag_compile(r"^Epoch\s*:\s*(\S+)")),
     _NameValue("release", re_tag_compile(r"^Release\s*:\s*(\S+)")),
     _NameValue("summary", re_tag_compile(r"^Summary\s*:\s*(.+)")),
+    _NameValue("description", re_tag_compile(r"^%description\s*(\S*)")),
+    _NameValue("changelog", re_tag_compile(r"^%changelog\s*(\S*)")),
     _NameValue("license", re_tag_compile(r"^License\s*:\s*(.+)")),
     _NameValue("group", re_tag_compile(r"^Group\s*:\s*(.+)")),
     _NameValue("url", re_tag_compile(r"^URL\s*:\s*(\S+)")),
     _NameValue("buildroot", re_tag_compile(r"^BuildRoot\s*:\s*(\S+)")),
-    _NameValue("buildarch", re_tag_compile(r"^BuildArch\s*:\s*(\S+)")),
+    _SplitValue("buildarch", re_tag_compile(r"^BuildArch\s*:\s*(\S+)")),
+    _SplitValue("excludearch", re_tag_compile(r"^ExcludeArch\s*:\s*(.+)")),
+    _SplitValue("exclusivearch", re_tag_compile(r"^ExclusiveArch\s*:\s*(.+)")),
     _ListAndDict("sources", re_tag_compile(r"^(Source\d*\s*):\s*(.+)")),
     _ListAndDict("patches", re_tag_compile(r"^(Patch\d*\s*):\s*(\S+)")),
     _List("build_requires", re_tag_compile(r"^BuildRequires\s*:\s*(.+)")),
@@ -189,6 +275,7 @@ _tags = [
     _List("packages", re.compile(r"^%package\s+(\S+)")),
     _MacroDef("define", re.compile(r"^%define\s+(\S+)\s+(\S+)")),
     _MacroDef("global", re.compile(r"^%global\s+(\S+)\s+(\S+)")),
+    _DummyMacroDef("dummy", re.compile(r"^%[a-z_]+\b.*$")),
 ]
 
 _tag_names = [tag.name for tag in _tags]
@@ -196,11 +283,20 @@ _tag_names = [tag.name for tag in _tags]
 _macro_pattern = re.compile(r"%{(\S+?)\}")
 
 
-def _parse(spec_obj, context, line):
+def _parse(spec_obj: "Spec", context: Dict[str, Any], line: str) -> Any:
     for tag in _tags:
         match = tag.test(line)
         if match:
+            if "multiline" in context:
+                context.pop("multiline", None)
             return tag.update(spec_obj, context, match, line)
+    if "multiline" in context:
+        target_obj = _Tag.current_target(spec_obj, context)
+        previous_txt = getattr(target_obj, context["multiline"], "")
+        if previous_txt is None:
+            previous_txt = ""
+        setattr(target_obj, context["multiline"], str(previous_txt) + line + os.linesep)
+
     return spec_obj, context
 
 
@@ -233,9 +329,12 @@ class Requirement:
 
     expr = re.compile(r"(.*?)\s+([<>]=?|=)\s+(\S+)")
 
-    def __init__(self, name):
+    def __init__(self, name: str) -> None:
         assert isinstance(name, str)
         self.line = name
+        self.name: str
+        self.operator: Optional[str]
+        self.version: Optional[str]
         match = Requirement.expr.match(name)
         if match:
             self.name = match.group(1)
@@ -246,8 +345,15 @@ class Requirement:
             self.operator = None
             self.version = None
 
+    def __eq__(self, o: object) -> bool:
+        if isinstance(o, str):
+            return self.line == o
+        if isinstance(o, Requirement):
+            return self.name == o.name and self.operator == o.operator and self.version == o.version
+        return False
+
     def __repr__(self):
-        return self.line
+        return f"Requirement('{self.line}')"
 
 
 class Package:
@@ -291,7 +397,7 @@ class Package:
 
     """
 
-    def __init__(self, name):
+    def __init__(self, name: str) -> None:
         assert isinstance(name, str)
 
         for tag in _tags:
@@ -303,30 +409,37 @@ class Package:
                 "provides",
             ]:
                 setattr(self, tag.name, tag.attr_type())
+            elif tag.name in [
+                "description",
+            ]:
+                setattr(self, tag.name, None)
 
         self.name = name
         self.is_subpackage = False
 
-    def __repr__(self):
-        return "Package('{}')".format(self.name)
+    def __repr__(self) -> str:
+        return f"Package('{self.name}')"
 
 
 class Spec:
     """Represents a single spec file."""
 
-    def __init__(self):
+    def __init__(self) -> None:
         for tag in _tags:
             if tag.attr_type is list:
                 setattr(self, tag.name, tag.attr_type())
             else:
                 setattr(self, tag.name, None)
 
-        self.sources_dict = {}
-        self.patches_dict = {}
-        self.macros = {}
+        self.sources_dict: Dict[str, str] = {}
+        self.patches_dict: Dict[str, str] = {}
+        self.macros: Dict[str, str] = {}
+
+        self.name: Optional[str]
+        self.packages: List[Package] = []
 
     @property
-    def packages_dict(self):
+    def packages_dict(self) -> Dict[str, Package]:
         """All packages in this RPM spec as a dictionary.
 
         You can access the individual packages by their package name, e.g.,
@@ -338,7 +451,7 @@ class Spec:
         return dict(zip([package.name for package in self.packages], self.packages))
 
     @staticmethod
-    def from_file(filename):
+    def from_file(filename: str) -> "Spec":
         """Creates a new Spec object from a given file.
 
         :param filename: The path to the spec file.
@@ -349,11 +462,11 @@ class Spec:
         with open(filename, "r", encoding="utf-8") as f:
             parse_context = {"current_subpackage": None}
             for line in f:
-                spec, parse_context = _parse(spec, parse_context, line)
+                spec, parse_context = _parse(spec, parse_context, line.rstrip())
         return spec
 
     @staticmethod
-    def from_string(string: str):
+    def from_string(string: str) -> "Spec":
         """Creates a new Spec object from a given string.
 
         :param string: The contents of a spec file.
@@ -367,20 +480,18 @@ class Spec:
         return spec
 
 
-def replace_macros(string, spec=None):
+def replace_macros(string: str, spec: Spec) -> str:
     """Replace all macros in given string with corresponding values.
 
     For example: a string '%{name}-%{version}.tar.gz' will be transformed to 'foo-2.0.tar.gz'.
 
-    :param string A string containing macros that you want to be replaced
-    :param spec An optional spec file. If given, definitions in that spec
-    file will be used to replace macros.
+    :param string A string containing macros that you want to be replaced.
+    :param spec A Spec object. Definitions in that spec file will be used to replace macros.
 
     :return A string where all macros in given input are substituted as good as possible.
 
     """
-    if spec:
-        assert isinstance(spec, Spec)
+    assert isinstance(spec, Spec)
 
     def _is_conditional(macro: str) -> bool:
         return macro.startswith("?") or macro.startswith("!")
@@ -393,6 +504,7 @@ def replace_macros(string, spec=None):
         raise Exception("Given string is not a conditional macro")
 
     def _macro_repl(match):
+        # pylint: disable=too-many-return-statements
         macro_name = match.group(1)
         if _is_conditional(macro_name) and spec:
             parts = macro_name[1:].split(sep=":", maxsplit=1)
